/**
  Physics simulator demo program.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2011-01-20 (Public Domain)
*/
#include "physics/world.h"
#include "ogl/glsl.h"

/* LAME! just include library bodies here, for easy linking */
#include "physics/world.cpp" 
#include "physics/config.cpp"
#include "ogl/glsl.cpp"

/* One simulated particle */
class particle {
public:
	vec4 color; // RGBA
	vec3 pos, vel; // in world coordinates-- meters
	vec3 netforce; // in newtons (quarter-pounds)
	
	void start_step(double dt) {
		/*
		 For stability, I'm using "Leapfrog" integration:
			http://en.wikipedia.org/wiki/Leapfrog_integration
		 Basically, we just update position, *then* calculate
		 net force before moving the velocity.
		*/
		pos+=dt*vel;
		
		/* Zero out net force, so springs can add to it */
		netforce=vec3(0.0);
	}
	
	void finish_step(double dt,const vec3 &target) {
	/* Newtonian motion */
		float mass=1.0; // kilograms (for this particle)
		vec3 acc=netforce/mass; 

		acc.z+=-9.8;  /* gravity */
		
		/* Hazard avoidance force */
		vec3 fear=-(target-pos);
		if (length(fear)<1.0) { // close enough to scare me! 
			acc+=normalize(fear)*1000.0;
		}
		
		/* "wind drag": fairly gentle */
		acc+=-1.0*vel;
		vel+=dt*acc;

		if (pos.z<0.0) { /* hit the ground */
			vel.z=0.6*fabs(vel.z);
			vel*=0.4; // ground drag
			pos.z=0.0;
		}
		if (pos.x<-2.0) { /* hit the left wall */
			vel.x=0.6*fabs(vel.x);
			pos.x=-2.0;
		}
	}
};

/* Random float between 0 and 1 */
float randf(void) {
	float f=rand();
	return f/RAND_MAX;
}

/* Random vector where x y and z vary between 0 and 1 */
vec3 randVec3(void) {
	return vec3(randf(),randf(),randf());
}

/* A 3D tetrahedron */
class tet {
public:
	int v[4]; /* each int is a vertex index */
	float restlength[6]; // length in meters of rest position of each edge
};

void bad(const char *err) {
	std::cout<<"Fatal file read error:  "<<err<<"\n";
	exit(1);
}

/* A particle system */
class particlesystem : public physics::object {
	std::vector<particle> p;
	std::vector<tet> tets;
	vec3 targetpoint;
	bool initRestState;
public:
	particlesystem(void) 
		:physics::object(0.01)  /* <- our timestep, in seconds */
	{
		initRestState=true;
		
		/* Read a TetGen set of tetrahedra
		      http://tetgen.berlios.de/
		*/
		std::ifstream f("model.1.ele");
		int ntet,npertet,nedge;
		f>>ntet>>npertet>>nedge;
		if (npertet!=4) bad("expected 4 vertices per tet");
		
		/* Read all the tets */
		for (int i=0;i<ntet;i++) {
			tet t;
			int tetno; f>>tetno; if (i!=tetno) bad("Unexpected tet number!");
			for (int j=0;j<npertet;j++)
				f>>t.v[j];
			tets.push_back(t);
		}
		if (!f) bad("I/O error reading .ele file");
		f.close();
		
		/* Read all the XYZ vertices */
		f.open("model.1.node");
		int nnode, coordpernode, val1pernode, val2pernode;
		f>>nnode>> coordpernode>> val1pernode>> val2pernode;
		for (int i=0;i<nnode;i++) {
			int nodeno; f>>nodeno; if (nodeno!=i) bad("Unexpected node number!");
			float x,y,z;
			f>>x>>y>>z;
			particle np;
			np.color=1.0*randVec3(); np.color.w=0.5; /* alpha */
			np.pos=vec3(x,z,-y)+vec3(2.0,0,1.3);
			np.vel=vec3(-8.0,0.0,5.0); // (randVec3()-vec3(0.5));
			p.push_back(np);
		}
		if (!f) bad("I/O error reading .node file");
		f.close();
	}
	
	void simulate(physics::library &lib) {
		float slowmo=0.1;
		float dt=timestep.dt*slowmo;
	
		// Start timesteps
		for (unsigned int i=0;i<p.size();i++)
		{
			p[i].start_step(dt);
			if (length(p[i].vel)>10000.0) {std::cout<<"Unstable at t="<<timestep.last_t<<"\n"; exit(1);}
		}
		
		// Loop over edges, add in net forces to particles
		for (unsigned int i=0;i<tets.size();i++) {
			int edgeno=0;
			for (int j=0;j<4;j++) 
			for (int k=j+1;k<4;k++, edgeno++) {
				int a=tets[i].v[j];
				int b=tets[i].v[k];
				
				vec3 dist=p[a].pos-p[b].pos;
				vec3 along=normalize(dist);
				float len=length(dist);
				
				/* Newtons of force per meter of displacement:
				    this should be pretty huge, in the thousands
				*/
				float springconst=20000.0;
				
				if (initRestState) tets[i].restlength[edgeno]=len-0.0001;
				
				/* Spring force is straight from Hooke's Law */
				float f=-springconst*(len-tets[i].restlength[edgeno]);
				
				p[a].netforce+=f*along;
				p[b].netforce-=f*along;
			}
		}
		initRestState=false;
		
		
		targetpoint=camera-camera_orient.z*3.0-camera_orient.y*1.0;
	
		// Finish timesteps
		for (unsigned int i=0;i<p.size();i++) {
			p[i].finish_step(dt,targetpoint);
		}
	}
	
	void draw(physics::library &lib) {
		
		/* Draw each of the nodes */
		glEnable(GL_POINT_SMOOTH);
		glPointSize(20.0);
		glBegin(GL_POINTS);
		for (unsigned int i=0;i<p.size();i++) {
			glColor4fv(p[i].color);
			glVertex3fv(p[i].pos);
			
			// Draw a "shadow"
			glColor4f(0.0f,0.0f,0.0f,0.1f);
			glVertex2fv(p[i].pos); /* 2fv means z==0 */
		}
		glEnd();
		
		/* Outline the tet edges in blue */
		glColor4f(0,0,1,1);
		glLineWidth(1.0);
		glBegin(GL_LINES);
		for (unsigned int i=0;i<tets.size();i++) {
			for (int j=0;j<4;j++) 
			for (int k=j+1;k<4;k++) {
				glVertex3fv(p[tets[i].v[j]].pos);
				glVertex3fv(p[tets[i].v[k]].pos);
			}
		}
		glEnd();
		
		/* Green sphere is our pusher */
		glLineWidth(2.0);
		for (int shadow=0;shadow<=1;shadow++) {
			glPushMatrix();
			glColor3f(0,1,0);
			if (shadow) {
				glScalef(1,1,0);
				glColor4f(0,0,0,0.3);
			}
			glTranslatef(VEC3_TO_XYZ(targetpoint));
			glutWireSphere(1.0,6,8);
			glPopMatrix();
		}
		
		/* Red quad represents the left bumper */
		glLineWidth(1.0);
		for (int outline=0;outline<=1;outline++) {
			glPushMatrix();
			glTranslatef(-2.0,0,+1);
			if (outline) glDepthMask(GL_TRUE); else glDepthMask(GL_FALSE);
			if (outline) glColor3f(0,0,0); else glColor4f(1,0,0,0.3);
			if (outline) glBegin(GL_LINE_LOOP); else glBegin(GL_QUADS);
				glVertex3f(0,-30,-4);glVertex3f(0,+30,-4);
				glVertex3f(0,+30,+4);glVertex3f(0,-30,+4);
			glEnd();
			glPopMatrix();
		}
	}
};


/* Called create a new simulation */
void physics_setup(physics::library &lib) {
	camera=vec3(0.0,-5.0,1.7);
	lib.world->add(new particlesystem());
	lib.background[0]=lib.background[1]=0.6;
	lib.background[2]=1.0; // light blue
}

/* Called every frame to draw everything */
void physics_draw_frame(physics::library &lib) {
	/* Point and line smoothing needs normal alpha blending */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	/* Draw opaque white ground plane */
	int groundsize=30;
	glEnable(GL_DEPTH_TEST);
	glColor3f(1,1,1);
	glBegin(GL_QUADS);
	glVertex2i(-groundsize,-groundsize); glVertex2i(+groundsize,-groundsize);
	glVertex2i(+groundsize,+groundsize); glVertex2i(-groundsize,+groundsize);
	glEnd();
	
	/* No depth while drawing ground grid */
	glDisable(GL_DEPTH_TEST);
	glColor4f(0.2,0.2,0.2,0.2);
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(1.0);
	if (camera.z>=0.0) { /* no grid if you're underneath ground */
		glBegin(GL_LINES);
		for (int grid=-groundsize;grid<=+groundsize;grid++) {
			glVertex3i(-groundsize,grid,0); glVertex3i(+groundsize,grid,0); 
			glVertex3i(grid,-groundsize,0); glVertex3i(grid,+groundsize,0); 
		}
		glEnd();
	}
	
	/* Do depth while drawing objects */
	glEnable(GL_DEPTH_TEST);
	lib.world->draw(lib);
}


